/*******************************************************************************/
/* Permission is hereby granted, free of charge, to any person or organization */
/* obtaining a copy of the software and accompanying documentation covered by  */
/* this license (the "Software") to use, reproduce, display, distribute,       */
/* execute, and transmit the Software, and to prepare derivative works of the  */
/* Software, and to permit third-parties to whom the Software is furnished to  */
/* do so, all subject to the following:                                        */
/*                                                                             */
/* The copyright notices in the Software and this entire statement, including  */
/* the above license grant, this restriction and the following disclaimer,     */
/* must be included in all copies of the Software, in whole or in part, and    */
/* all derivative works of the Software, unless such copies or derivative      */
/* works are solely in the form of machine-executable object code generated by */
/* a source language processor.                                                */
/*                                                                             */
/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  */
/* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    */
/* FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT   */
/* SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE   */
/* FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE, */
/* ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER */
/* DEALINGS IN THE SOFTWARE.                                                   */
/*******************************************************************************/

#include <lfp/signal.h>

#if defined(__OpenBSD__) || defined(__NetBSD__) || defined(__FreeBSD__) || defined(__DragonFly__)
# include <sys/signalvar.h>
#endif

DSO_PUBLIC lfp_sighandler_t
lfp_sig_dfl (void)
{
  return SIG_DFL;
}

DSO_PUBLIC lfp_sighandler_t
lfp_sig_err (void)
{
  return SIG_ERR;
}

DSO_PUBLIC lfp_sighandler_t
lfp_sig_hold (void)
{
  return SIG_HOLD;
}

DSO_PUBLIC lfp_sighandler_t
lfp_sig_ign (void)
{
  return SIG_IGN;
}

DSO_PUBLIC int
lfp_sigrtmin (void)
{
#if defined(SIGRTMIN)
  return SIGRTMIN;
#else
  SYSERR(ENOSYS);
#endif // SIGRTMIN
}

DSO_PUBLIC int
lfp_sigrtmax (void)
{
#if defined(SIGRTMAX)
  return SIGRTMAX;
#else
  SYSERR(ENOSYS);
#endif // SIGRTMAX
}
